module GROESTL_CORE(
   clk, rst_n,
   init, start, busy, Ld_msg,
   hash, idata, EOM
);

//{{{ input & output
input clk;
input rst_n;

input init;
input start;

input [15:0] idata;

input EOM;
input Ld_msg;

output [255:0] hash;
output busy;
//}}}

//{{{ register & wire
reg [511:0] stateP;
reg [511:0] stateQ;
reg [511:0] hash_r;

reg [4:0] round;

reg EN;

reg init_r;

wire [511:0] cv;
wire [511:0] pout;
wire [511:0] qout;

wire [511:0] pin;
wire [511:0] qin;

//}}}

//{{{ core ctrl
//busy
assign busy = (start | EN)? 1:0;

//{{{ EN
always @(posedge clk or negedge rst_n) begin
   if (~rst_n) EN <= 0;
   else if (start) EN <= 1;
   else if (round == 5'h9) begin
      if (EOM) EN <= 1;
      else EN <= 0;
   end
   else if (round == 5'h19) EN <= 0;
   else EN <= EN;
end
//}}}

//{{{ round count
always @(posedge clk or negedge rst_n) begin
   if (~rst_n) round <= 0;
   else if (EN) begin
      if (round == 5'h19) round <= 0;
      else if (round == 5'h9) begin 
         if (EOM) round <= 5'h10;
         else round <= 0;
      end
      else round <= round + 1;
   end
   else round <= round;
end
//}}}

//{{{init_r
always @(posedge clk or negedge rst_n) begin
   if (~rst_n) init_r <= 0;
   else init_r <= init;
end
//}}}

//}}}

//{{{ stateQ 
always @(posedge clk or negedge rst_n) begin
   if (~rst_n) stateQ <= 0;
   else if (Ld_msg) stateQ <= {stateQ[496:0],idata};
   else if (EN) stateQ <= qout;
   else stateQ <= stateQ;
end
//}}}

//{{{ stateP 
always @(posedge clk or negedge rst_n) begin
   if (~rst_n) stateP <= 0;
   else if (EN) stateP <= pout;
   else stateP <= stateP;
end
//}}}

//{{{ hash_r
assign cv = hash_r ^ pout;
assign hash = hash_r[255:0];
always @(posedge clk or negedge rst_n) begin
   if (~rst_n) hash_r <= 0;
   else if (init_r) hash_r <= {496'h0,16'h0100};
   else if (round == 5'h9) hash_r <= pout ^ qout ^ hash_r;
   else if (round == 5'h19) hash_r <= cv;
   else hash_r <= hash_r;
end
//}}}
   
assign pin = (round == 5'h0)? hash_r ^ stateQ : 
   (round == 5'h10)? hash_r : stateP;

assign qin = stateQ;

Permute P(pin, pout, {4'h0,round[3:0]},1'b1);
Permute Q(qin, qout, {4'h0,round[3:0]},1'b0);

endmodule
